
#
# IBM Confidential
# OCO Source Materials
# Rational ClearCase
# (C) Copyright IBM Corp. 1990, 2010.  All Rights Reserved.
#
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#

BEGIN { 
    if ($ENV{CLEARCASE_PERLLIB}){
	for (@INC){$_ = "" unless /clearcase/i or /\.\.[\\\/]lib/ or /atria/i or /rational/i} 
    }
}

use HTML::Normalize;
use strict;
use Config;

my $UNIX = 0;
if ($Config::Config{'osname'} ne 'MSWin32') {
    $UNIX = 1;
}

my $perlname = "ccperl ";
my $sep = "\\";
if ($UNIX) {
    $perlname = "";
    $sep = "/";
}

my $logfile = "";
my $logdir = "";
my $tmpdir = "###THIS_IS_AN_INVALID_DIR###";
my $arg;
my $flagsdone;
my %filemap; # Maps filenames to tempfile names
my @filenames; # Need this to keep filenames in order
my $basename = ""; # Filename for base file if specified
my %seenfiles; # Keeps track of filenames we've seen on the cmd line
my @fnamed_files; # Files specified on the command line with -fname
my @args; # Args passed to the diff/merge tool we invoke
my $operation; # Which type manager op we're doing: compare, xcompare,
               #                                    merge, xmerge
my $default_extension = "";   # temp files with no extension provided will use this extension
my $debug;

$debug = "no";
$debug = "yes" if $ENV{HTMLNORMDIFF_DEBUG};
my $writeable;

# Split using either separator, since both are allowed by
# ClearCase on NT.
my $seppattern = "\[\\\/\\\\\]";
my $parse_error = "no";
my $real_tmpdir = $ENV{"TMP"};
if ($UNIX && ($real_tmpdir =~ /^\s*$/)) {
    $real_tmpdir = '/tmp' ;
}

# Clear log file
if ("$debug" eq "yes") {

    $logdir = $real_tmpdir;
    if ($logdir eq "") {
        $logdir = ".";
    }

    $logfile = "htdiffmrg.log";
    $logfile = $logdir . $sep . $logfile;
    open LOG, ">$logfile" or die "Can't open \"$logfile\" for writing.\n";
}

# sub add_argfile
# Create a temp file corresponding to the file passed as argument
# and add its name to the input arg list (module-global).
# Keep track of the relationship between the two files
# in module-global filemap.
sub add_argfile
{
    my $filename = shift;
    print LOG "Filename: $filename\n" if ("$debug" eq "yes");

    if (! $UNIX) {
        $filename =~ s/\//\\/g;
    }

    my $filenamecounter;
    my $uniqname;
    my @pathcomps = split("$seppattern", $filename);
    my $tempname = $pathcomps[$#pathcomps];
    my $ext;
    if ($tempname =~ s/\.(\w+$)//) { # Strip the extension
	$ext = $1;
        # if it's a cleartext object from a msdos-text-mode view, whack the
        # extension.
        if ("$ext" eq "msdos") {
            print LOG "found extension \".msdos\": assuming msdos-text-mode cleartext object, deleting extension.\n" if ("$debug" eq "yes");
            $ext = "";
        } else {
            if ("$default_extension" eq "") { # use this extension as default
                print LOG "setting default extension to \"$ext\"\n" if ("$debug" eq "yes");
                $default_extension = $ext;
            }
        }   
    }

    # If we've seen the filename, $tempname, add a uniqifier
    $uniqname = $tempname;
    while ($seenfiles{$uniqname}) {
        $uniqname = $tempname . '_' . ++$filenamecounter;
    }
    $seenfiles{$uniqname} = 1;
    $tempname = $tmpdir . $sep . $uniqname . "_tmp_$$";
    if ($ext) {
	$tempname .= '.' . $ext;
        print LOG "extension supplied for this argument. Tacking \"$ext\" onto temp filename\n" if ("$debug" eq "yes");
    } else {
        # there was no extension. Use the default extension if we have one, or ".htm" if we don't
        # IE does wierd things with a 0-length file if the extension isn't "html"
        if ("$default_extension" eq "") {
            $tempname .= '.' . "htm";
            print LOG "extension NOT supplied for this argument and no default extension active. Hardcoding \"\.htm\" onto temp filename\n" if ("$debug" eq "yes");
        } else {
            $tempname .= '.' . $default_extension;
            print LOG "extension NOT supplied for this argument and a default extension IS active. Tacking  \"$default_extension\" onto temp filename\n" if ("$debug" eq "yes");
        }
    }

    # directory for temp file is added later once the fnames are known

    print LOG "  Temporary: $tempname\n" if ("$debug" eq "yes");

    push (@filenames, $filename);
    $filemap{$filename} = $tempname;
}

print LOG "htmlnormdiff.pl: args: " . join(' ', @ARGV) . "\n" if ("$debug" eq "yes");
if ($#ARGV < 2) {
    die "Usage: ${perlname}htmlnormdiff.pl <operation> <cleardiffmrg options>\nARGV was " . join(' ', @ARGV);
}

$operation = shift(@ARGV);
if ($operation !~ /^x?(merge|compare)$/) {
print LOG <<EOM;
${perlname}htmlnormdiff.pl: unknown operation \"$operation\".
  Note that htmlnormdiff is intended to be executed only by
  a ClearCase type manager when doing a diff or merge operation.
EOM
    die("Exiting diff/merge");
}

# Replace filenames with tempfile names on command line
while ($arg = shift(@ARGV)) {

    print LOG "Processing ARGV $arg\n" if ("$debug" eq "yes");
    
    #if the arg is -column then we need push it, then shift to get
    #its parameter and push that as well.

    if ($arg =~ /^-col/) {
	push(@args, $arg);
	$arg = shift(@ARGV);
	push(@args, $arg);
    }

    # Flagged args always come before filenames, which are final args
    elsif ($flagsdone || $arg !~ /^-/) {
	add_argfile($arg);
	$flagsdone = 1;
    }

    else {
        # push it if it's anything other than -base
        if (! ($arg =~ /^-bas/) ) {
	    push(@args, $arg);
        }
	# If it's a -base, -out, or -fname argument, the next is a filename
	if ($arg =~ /^-(out|bas|fna)/) {
	    $arg = shift(@ARGV);

            print LOG "Processing ARGV $arg\n" if ("$debug" eq "yes");

	    if ($1 eq 'bas') {
		# Base file
		add_argfile($arg);
		$basename = $arg;
		# push later after name substitution
	    }
	    else {
		if ($1 eq 'fna') {
		    # Don't specify the fname later, because caller did
                    print LOG "pushing -fname argument $arg\n" if ("$debug" eq "yes");
		    if (! $UNIX) {
        		$arg =~ s/\//\\/g;
    		    }
		    push (@fnamed_files, $arg);

                    # grab extension of fnamed arg to set default extension if not
                    # already set
                    my $ext = "";
                    $ext = $arg;
                    $ext =~ s/\@\@.*$//; # strip of everything after @@
                    $ext =~ s/^.*\.//;
                    if ("$ext" ne "") {
                        print LOG "fname extension is $ext\n" if ("$debug" eq "yes");
                        if ("$default_extension" eq "") { # use this extension as default
                            print LOG "setting default extension to \"$ext\"\n" if ("$debug" eq "yes");
                            $default_extension = $ext;
                        }
                    }
		}
		push(@args, "\"" . $arg . "\"");
	    }
	}
    }
}

# Add the -fname arguments for display
my @fnames = @filenames;
my $dirpath = "";
my $attrval = "";
my $will_normalize = "yes";
while ($arg = shift(@fnames)) {

    my $thisfname = shift(@fnamed_files);
    if (!$thisfname) { # If command line didn't specify -fname
	push (@args, ('-fname', "\"" . $arg . "\""));
        $dirpath = $arg; # take dir off of argument
    } else {
        $dirpath = $thisfname # take it from the -fname argument
    }

    # make the temporary file go into the directory of the fname/arg
    # strip any @@ from fname
    $dirpath =~ s/\@\@.*$//;
    my $fname_element = $dirpath;
    # now strip leaf leaving the fname directory
    $dirpath =~ s/[^\\\/]*$//;
    print LOG "\$dirpath is $dirpath\n" if ("$debug" eq "yes");
    $writeable = &is_writeable($dirpath);
    # check for name with no separator
    if (! ($dirpath =~ /[\\\/]/)) {
      if($writeable){
	  $filemap{$arg} =~  s/$tmpdir/\./;
      } else {
	  $filemap{$arg} =~  s/$tmpdir/$real_tmpdir/;
      }
      #$filemap{$arg} =~  s/$seppattern//;
      print LOG "changing temp for $arg to $filemap{$arg}\n" if ("$debug" eq "yes");
    } else {
      my $curtempname = $filemap{$arg};
      if($writeable){
          $curtempname =~ s/^$tmpdir//;
          $curtempname =~ s/^[\\\/]//;
          my $newtemp = $dirpath . $curtempname;
          if ($filemap{$arg} =~ /$tmpdir/) {
	      print LOG "changing temp for $arg to $newtemp\n" if ("$debug" eq "yes");
	      $filemap{$arg} =  $newtemp;
	  } else {
	      print LOG "already changed temp for $arg to $filemap{$arg}\n" if ("$debug" eq "yes");
	  }
      } else {
	  $filemap{$arg} =~ s/$tmpdir/$real_tmpdir/;
      }
    }
    
    # check for the presense of the normalize_html attribute on this fname
    # if it's there for all fnames, we will normalize
    
    # add @@ to name
    $fname_element = $fname_element . "\@\@";
    print LOG "checking attribute normalize_html on $fname_element\n" if ("$debug" eq "yes");

    $attrval = "";
    $attrval = `cleartool desc -fmt "%[normalize_html]Sa" \"$fname_element\"`;
    print LOG "attrval is $attrval\n" if ("$debug" eq "yes");
    if ("$attrval" =~ /\(\"yes\"\)/) {
        print LOG "force normalization attribute is on\n" if ("$debug" eq "yes");
    }
    if (("$attrval" =~ /\(\"no\"\)/) || ("$attrval" eq "")) {
        print LOG "force normalization OFF\n" if ("$debug" eq "yes");
        $will_normalize = "no";
    }

}

@fnames = @filenames;
my $firstbase = "yes";
while ($arg = shift(@fnames)) {
    if (($arg ne $basename) || ("$firstbase" eq "no")) {
	push (@args, "\"" . $filemap{$arg} . "\"");
    } else {
        $firstbase = "no";
        push (@args, "-base " . "\"" . $filemap{$arg} . "\"");
    }
}

# Process files
$parse_error = "no";
foreach (keys(%filemap)) {

    # if the file being diff'ed is an unloaded element in a snapshot
    # view, the directory for it may not exist. So mkdir -p the directory
    # if necessary

    my $tmpfile = $filemap{$_};
    my $pattern = "\[\^" . "\\" . $sep . "\]\*\$";
    my $tmpdir = $tmpfile;
    $tmpdir =~ s/$pattern//;
    $tmpdir =~ s/[\\\/]$//;
    if (! -d "$tmpdir") {
      print LOG "directory for tempfile $tmpfile does not exist.\n" if ("$debug" eq "yes");
      my @pathcomps = split(/[\/\\]/, $tmpdir);
      my $dircomp = "";
      my $dir = $dircomp;
      while($dircomp = shift(@pathcomps)) {
          $dir = $dir . $dircomp . $sep;
          my $cdir  = $dir;
          if (! ($cdir =~ /^[a-zA-Z]\:$seppattern$/)) {
              $cdir =~ s/.$//;
          }
          if (! -d "$cdir") {
            print LOG "dir doesn't exist, doing mkdir\n" if ("$debug" eq "yes");
            mkdir("$cdir", 777) || die("error in mkdir\n");
          } else {
            ;
          }
      }
    }

    open FILE, ">$filemap{$_}" or 
		      die ("Can't create temporary file \"$filemap{$_}\"\n");
    close FILE;

    if ("$will_normalize" eq "no") {

        print LOG "Copying HTML from \"$_\" to \"$filemap{$_}\"\n" if ("$debug" eq "yes");

        open(SOURCE_FILE, "<$_") || die("error opening HTML source file");
        open FILE, ">>$filemap{$_}" or 
		      die ("Can't open temporary file \"$filemap{$_}\"\n");
        while(<SOURCE_FILE>) {
            print FILE $_;
        }
        close SOURCE_FILE;
    } else {

        print LOG "Normalizing HTML in \"$_\" to \"$filemap{$_}\"\n" if ("$debug" eq "yes");

        close LOG if ("$debug" eq "yes");

        my $html = HTML::Normalize->new();
        if (!defined (eval{$html->parse_file($_, $filemap{$_}, $logfile)})) {

            open LOG, ">>$logfile" or die "Can't open \"$logfile\" for writing.\n" if ("$debug" eq "yes");

            if ("$@" =~ /NO_MATCHING_GENERATOR/) { # no generator meta-tag
               print LOG "no matching generator tag in this file\n" if ("$debug" eq "yes");
           } else {
                $parse_error = "yes";
	        print $@;
	        print "Stopped processing $_\n";
                print "parse results can be seen in file $filemap{$_}\n";
                system("clearprompt proceed -type error -mask proceed -prompt \"HTML Parsing error. Parser returned '$@'. Parse results can be seen in file \"$filemap{$_}\"\""); 
            }
        } else {
            open LOG, ">>$logfile" or die "Can't open \"$logfile\" for writing.\n" if ("$debug" eq "yes");
        }
    }

    close FILE;

}

# Execute the tool
my @syscmd;
if ($operation =~ /^x/) {
    if ($UNIX) {
	$syscmd[0] = "xcleardiff";
    }
    else {
	$syscmd[0] = "cleardiffmrg";
    }
    $syscmd[1] = "-html";
}
else {
    $syscmd[0] = "cleardiff";
}

push(@syscmd, join(' ', @args));

my $status = 0;

if ("$parse_error" eq "no") {
    print LOG "Invoking " . join(' ', @syscmd) . "\n" if ("$debug" eq "yes");
    $status = system(join(' ', @syscmd));

    $status = $status >> 8;
} else {
    $status = 2;
}

# Remove temp files
if (($parse_error eq "no") && ("$debug" eq "no")) {
    @fnames = @filenames;
    while ($arg = shift(@fnames)) {
        print LOG "deleting $filemap{$arg}\n" if ("$debug" eq "yes");
	unlink $filemap{$arg};
    }
}
    

print LOG "Exiting with status $status\n" if ("$debug" eq "yes");

close LOG;

exit ($status);


sub is_writeable{
    my $dirpath = shift;
    my $fn;        # File name for writeable test
    # Find out if directory is writeable.  -w switch applies only to files
    $fn = $dirpath.'AFile.'.$$;
    if ( open FN, ">$fn" ){
	print LOG "$dirpath - $fn is writeable.\n" if ("$debug" eq "yes");
	close FN;
	unlink $fn;
	return 1;
    } else {
	print LOG "Warning: $dirpath - $fn is unwriteable.  Using $real_tmpdir\n" if ("$debug" eq "yes");
	return 0;
    }
}
